from uuid import UUID

from marshmallow import fields, post_dump, post_load, pre_dump, Schema, validate
from sqlalchemy.ext.hybrid import hybrid_property

from stumpf.diva_models import ACTIVITY_TYPES, OCCLUSION_TYPES, \
    SOURCE_TYPES, TRACK_TYPES
from stumpf.models import Fact, FactSchema


# Sub-models stored in the Fact table specific to diva
class TrackDetectionFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:track_detection'
    }

    @hybrid_property
    def top(self):
        return self.fact['top']

    @hybrid_property
    def bottom(self):
        return self.fact['bottom']

    @hybrid_property
    def left(self):
        return self.fact['left']

    @hybrid_property
    def right(self):
        return self.fact['right']

    @hybrid_property
    def source(self):
        return self.fact['source']

    @hybrid_property
    def activity(self):
        return self.fact['activity']

    @hybrid_property
    def occlusion(self):
        return self.fact['occlusion']

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def frame(self):
        return self.frame_range.lower


class TrackFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:track'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def detections(self):
        return list(TrackDetectionFact.query.filter(
            TrackDetectionFact.fact['track_id'].astext == str(self.id)
        ))


class ActivityFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:activity'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def tracks(self):
        return list(TrackFact.query.filter(
            TrackFact.fact['activity_id'].astext == str(self.id)
        ))


class GladiatorFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'gladiator:answer'
    }

    @hybrid_property
    def good(self):
        return self.fact['good']

    @hybrid_property
    def percent_good(self):
        return self.fact['good_count']

    @hybrid_property
    def reasons(self):
        return self.fact['reasons']

    @hybrid_property
    def activity(self):
        return ActivityFact.query.get(self.fact['activity_id'])


# Schemas for the jsonb portion of the fact model.
class GrozperFactPropertySchema(Schema):

    # Because we deserialize into a jsonb column, we need uuids to be
    # represented as strings.
    @post_load
    def convert_id_to_string(self, data):
        for key, value in data.items():
            if isinstance(value, UUID):
                data[key] = str(value)


class TrackDetectionFactPropertySchema(GrozperFactPropertySchema):
    track_id = fields.UUID(required=True, load_only=True)  # reference to a track fact

    source = fields.Str(required=True, validate=validate.OneOf(SOURCE_TYPES))
    occlussion = fields.Str(missing='none', validate=validate.OneOf(OCCLUSION_TYPES))
    left = fields.Int(required=True)
    right = fields.Int(required=True)
    top = fields.Int(required=True)
    bottom = fields.Int(required=True)

    # The type and activity is duplicated from the TrackFactSchema to help in the very common
    # case of querying detections by activity and track type.
    activity_id = fields.UUID(required=True, load_only=True)  # reference to an activity fact
    type = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES), load_only=True)


class TrackFactPropertySchema(GrozperFactPropertySchema):
    type = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES),
                      attribute='type_')
    activity_id = fields.UUID(required=True, load_only=True)  # reference to another fact
    detections = fields.List(
        fields.Nested(TrackDetectionFactPropertySchema), dump_only=True)


class ActivityFactPropertySchema(GrozperFactPropertySchema):
    type = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    tracks = fields.List(
        fields.Nested(TrackFactPropertySchema), dump_only=True)


# Schemas for the full fact models.  These derive from the base FactSchema
# and perform additional mutations to the fact serialization for compatibility
# with existing applications.
class GrozperFactSchema(FactSchema):
    @post_dump(pass_original=True)
    def dump_as_fact_property(self, data, original_data):
        factSchema = self.__fact_property_schema__()
        fact = factSchema.dump(original_data.fact)
        fact['id'] = data['id']
        fact['frame_range'] = [
            original_data.frame_range.lower,
            original_data.frame_range.upper - 1
        ]
        return fact


class TrackDetectionFactSchema(GrozperFactSchema):
    __fact_property_schema__ = TrackDetectionFactPropertySchema
    fact = fields.Nested(TrackDetectionFactPropertySchema, load_only=True)

    @post_dump
    def set_bounds_list(self, data):
        """Add a `bounds` key for compatibility with gladiator."""
        data['bounds'] = [data['left'], data['top'], data['right'], data['bottom']]
        return data

    @post_dump
    def set_frame_number(self, data):
        """Set the frame number from the top-level attribute."""
        data['frame'] = data['frame_range'][0]
        return data


class TrackFactSchema(GrozperFactSchema):
    __fact_property_schema__ = TrackFactPropertySchema
    fact = fields.Nested(TrackFactPropertySchema, load_only=True)

    @pre_dump
    def set_detections_list(self, data):
        data.fact['detections'] = getattr(data, 'detections', [])
        return data


class ActivityFactSchema(GrozperFactSchema):
    __fact_property_schema__ = ActivityFactPropertySchema
    fact = fields.Nested(ActivityFactPropertySchema, load_only=True)

    @pre_dump
    def set_detections_list(self, data):
        data.fact['tracks'] = getattr(data, 'tracks', [])
        return data

    @post_dump
    def set_compatible_frame_bounds(self, data):
        data['activity'] = {
            'start_frame': data['frame_range'][0],
            'end_frame': data['frame_range'][1]
        }
        return data
